from pdb import set_trace as st

from functools import reduce
from pathlib import Path
import os
import sys
from typing import Any, Dict, Union, Hashable, Optional, List

import numpy as np
from plotly import tools
from plotly.offline import init_notebook_mode, iplot
import plotly.graph_objs as go
from tensorboard.backend.event_processing.event_accumulator import \
    EventAccumulator, ScalarEvent
from tqdm import tqdm

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

import logging
logging.getLogger('tensorflow').addFilter(lambda x: 0)

BN_TYPES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def freeze_to(module: nn.Module,
              n: int,
              freeze_bn: bool = False) -> None:
    layers = list(module.children())
    for l in layers[:n]:
        for module in flatten_layer(l):
            if freeze_bn or not isinstance(module, BN_TYPES):
                set_grad(module, requires_grad=False)

    for l in layers[n:]:
        for module in flatten_layer(l):
            set_grad(module, requires_grad=True)


def freeze(module: nn.Module,
           freeze_bn: bool = False) -> None:
    freeze_to(module=module, n=-1, freeze_bn=freeze_bn)


def unfreeze(module: nn.Module) -> None:
    layers = list(module.children())
    for l in layers:
        for module in flatten_layer(l):
            set_grad(module, requires_grad=True)


def set_grad(module: nn.Module, requires_grad: bool) -> None:
    for param in module.parameters():
        param.requires_grad = requires_grad


# https://github.com/fastai/fastai/blob/6778fd518e95ea8e1ce1e31a2f96590ee254542c/fastai/torch_core.py#L157
class ParameterModule(nn.Module):
    """Register a lone parameter `p` in a module."""
    def __init__(self, p: nn.Parameter):
        super().__init__()
        self.val = p

    def forward(self, x): return x


# https://github.com/fastai/fastai/blob/6778fd518e95ea8e1ce1e31a2f96590ee254542c/fastai/torch_core.py#L149
def children_and_parameters(m: nn.Module):
    """Return the children of `m` and its direct parameters not registered in modules."""
    children = list(m.children())
    children_p = sum([[id(p) for p in c.parameters()] for c in m.children()],
                     [])
    for p in m.parameters():
        if id(p) not in children_p:
            st()
            children.append(ParameterModule(p))
    return children


def flatten_layer(layer: nn.Module) -> List[nn.Module]:
    if len(list(layer.children())):
        layers = []
        for children in children_and_parameters(layer):
            layers += flatten_layer(children)
        return layers
    else:
        return [layer]


def to_numpy(data: torch.Tensor) -> np.ndarray:
    return data.detach().cpu().numpy()


def exp_weight_average(curr_val: Union[float, torch.Tensor],
                       prev_val: float,
                       alpha: float = 0.9) -> float:
    if isinstance(curr_val, torch.Tensor):
        curr_val = to_numpy(curr_val)
    return float(alpha * prev_val + (1 - alpha) * curr_val)


def get_pbar(dataloader: DataLoader,
             description: str) -> tqdm:

    pbar = tqdm(
        total=len(dataloader),
        leave=True,
        ncols=0,
        desc=description,
        file=sys.stdout)

    return pbar


def extend_postfix(postfix: str, dct: Dict) -> str:
    if postfix is None:
        postfix = ""
    postfixes = [postfix] + [f"{k}={v:.4f}" for k, v in dct.items()]
    return ", ".join(postfixes)


def get_opt_lr(opt: torch.optim.Optimizer) -> float:
    lrs = [pg["lr"] for pg in opt.param_groups]
    res = reduce(lambda x, y: x + y, lrs) / len(lrs)
    return res


class DotDict(dict):
    """
    Example:
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr: str) -> Any:
        return self.get(attr)

    def __setattr__(self, key: Hashable, value: Any) -> Any:
        self.__setitem__(key, value)

    def __setitem__(self, key: Hashable, value: Any) -> Any:
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item: str) -> None:
        self.__delitem__(item)

    def __delitem__(self, key: str) -> None:
        super().__delitem__(key)
        del self.__dict__[key]


def load_state_dict(model: torch.nn.Module,
                    state_dict: Dict,
                    skip_wrong_shape: bool = False):
    model_state_dict = model.state_dict()

    for key in state_dict:
        if key in model_state_dict:
            if model_state_dict[key].shape == state_dict[key].shape:
                model_state_dict[key] = state_dict[key]
            elif not skip_wrong_shape:
                m = f"Shapes of the '{key}' parameters do not match: " \
                    f"{model_state_dict[key].shape} vs {state_dict[key].shape}"
                raise Exception(m)

    model.load_state_dict(model_state_dict)


def get_tensorboard_scalars(logdir: str,
                            metrics: Optional[List[str]],
                            step: str) -> Dict[str, List]:
    event_acc = EventAccumulator(str(logdir))
    event_acc.Reload()

    if metrics is not None:
        scalar_names = [n for n in event_acc.Tags()["scalars"] if step in n
                        and any(m in n for m in metrics)]
    else:
        scalar_names = [n for n in event_acc.Tags()["scalars"] if step in n]

    scalars = {sn: event_acc.Scalars(sn) for sn in scalar_names}
    return scalars


def get_scatter(scalars: Dict[str, ScalarEvent],
                name: str,
                prefix: str) -> go.Scatter:
    xs = [s.step for s in scalars[name]]
    ys = [s.value for s in scalars[name]]

    return go.Scatter(x=xs, y=ys, name=prefix+name)


def plot_tensorboard_log(logdir: Union[str, Path],
                         step: Optional[str] = "batch",
                         metrics: Optional[List[str]] = None,
                         height: Optional[int] = None,
                         width: Optional[int] = None) -> None:
    init_notebook_mode(connected=True)
    logdir = Path(logdir)

    train_scalars = get_tensorboard_scalars(logdir/"train", metrics, step)
    val_scalars = get_tensorboard_scalars(logdir/"val", metrics, step)

    if height is not None:
        height = height // len(train_scalars)

    for m in train_scalars:
        tm = get_scatter(train_scalars, m, prefix="train/")
        try:
            vm = get_scatter(val_scalars, m, prefix="val/")
            data = [tm, vm]
        except:
            data = [tm]
        layout = go.Layout(title=m,
                           height=height,
                           width=width,
                           yaxis=dict(hoverformat=".6f"))
        iplot(go.Figure(data=data, layout=layout))



from pdb import set_trace

from collections import defaultdict, namedtuple
from pathlib import Path
import shutil
from typing import Any, Callable, Dict, Tuple, Type, List, Optional, Union
import warnings

try:
    from apex import amp
except:
    pass  # warning message appears in keker.py module, no needs to be here

import numpy as np
import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau, _LRScheduler
from tensorboardX import SummaryWriter
from .utils import get_opt_lr, get_pbar, DotDict, \
    exp_weight_average, extend_postfix, to_numpy


class Callback:
    """
    Abstract base class used to build new callbacks.
    """
    def on_batch_begin(self, i: int, state: DotDict) -> None:
        pass

    def on_batch_end(self, i: int, state: DotDict) -> None:
        pass

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        pass

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        pass

    def on_train_begin(self, state: DotDict) -> None:
        pass

    def on_train_end(self, state: DotDict) -> None:
        pass


class Callbacks:
    def __init__(self, callbacks: Union[List, Any]) -> None:
        if isinstance(callbacks, Callbacks):
            self.callbacks = callbacks.callbacks
        if isinstance(callbacks, list):
            self.callbacks = callbacks
        else:
            self.callbacks = []

    def on_batch_begin(self, i: int, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_batch_begin(i, state)

    def on_batch_end(self, i: int, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_batch_end(i, state)

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_epoch_begin(epoch, epochs, state)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_epoch_end(epoch, state)

    def on_train_begin(self, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_train_begin(state)

    def on_train_end(self, state: DotDict) -> None:
        for cb in self.callbacks:
            cb.on_train_end(state)


class LRUpdater(Callback):
    """Basic class that all Lr updaters inherit from"""
    def __init__(self, init_lr: float) -> None:
        self.init_lr = init_lr

    def calc_lr(self) -> float:
        raise NotImplementedError

    def calc_momentum(self) -> float:
        raise NotImplementedError

    def update_lr(self,
                  optimizer: torch.optim.Optimizer) -> float:
        new_lr = self.calc_lr()
        for pg in optimizer.param_groups:
            pg["lr"] = new_lr
        return new_lr

    def update_momentum(self,
                        optimizer: torch.optim.Optimizer) -> float:
        new_momentum = self.calc_momentum()
        if "betas" in optimizer.param_groups[0]:
            for pg in optimizer.param_groups:
                pg["betas"] = (new_momentum, pg["betas"][1])
        else:
            for pg in optimizer.param_groups:
                pg["momentum"] = new_momentum
        return new_momentum

    def on_batch_begin(self, i: int, state: DotDict) -> None:
        if state.core.mode == "train":
            self.update_lr(state.core.opt)
            self.update_momentum(state.core.opt)


class OneCycleLR(LRUpdater):
    """
    An learning rate updater
        that implements the CircularLearningRate (CLR) scheme.
    Learning rate is increased then decreased linearly.
    Inspired by
    https://github.com/fastai/fastai/blob/master/fastai/callbacks/one_cycle.py
    """
    def __init__(self,
                 max_lr: float,
                 cycle_len: int,
                 len_loader: int,
                 momentum_range: Tuple[float, float],
                 div_factor: float,
                 increase_fraction: float) -> None:
        super().__init__(max_lr)
        self.cycle_len = cycle_len
        self.momentum_range = momentum_range
        self.div_factor = div_factor
        self.increase_fraction = increase_fraction
        self.len_loader = len_loader
        self.total_iter = None
        self.cycle_iter = 0
        # point in iterations for starting lr decreasing
        self.cut_point = None

    def on_train_begin(self, state: DotDict) -> None:
        self.total_iter = self.len_loader * self.cycle_len - 1
        self.cut_point = int(self.total_iter * self.increase_fraction)

    def calc_lr(self) -> float:
        # calculate percent for learning rate change
        if self.cycle_iter <= self.cut_point:
            percent = self.cycle_iter / self.cut_point

        else:
            percent = 1 - (self.cycle_iter - self.cut_point) / (
                    self.total_iter - self.cut_point)

        res = self.init_lr * (1 + percent * (self.div_factor - 1)) / self.div_factor

        return res

    def calc_momentum(self) -> float:
        if self.cycle_iter <= self.cut_point:
            percent = 1 - self.cycle_iter / self.cut_point

        else:
            percent = (self.cycle_iter - self.cut_point) / (self.total_iter - self.cut_point)
        res = self.momentum_range[1] + percent * (
                self.momentum_range[0] - self.momentum_range[1])
        return res

    def on_batch_begin(self, i: int, state: DotDict) -> None:
        super().on_batch_begin(i, state)
        if state.core.mode == "train":
            self.cycle_iter += 1


class LRFinder(LRUpdater):
    """
    Helps you find an optimal learning rate for a model,
        as per suggetion of 2015 CLR paper.
    Learning rate is increased in linear or log scale, depending on user input.

    https://sgugger.github.io/how-do-you-find-a-good-learning-rate.html
    """

    def __init__(self,
                 final_lr: float,
                 n_steps: int,
                 init_lr: float = 1e-6) -> None:
        super().__init__(init_lr)
        self.final_lr = final_lr
        self.n_steps = max(1, n_steps - 1)
        self.n = 0

    def calc_lr(self) -> float:
        res = self.init_lr * (self.final_lr / self.init_lr) ** \
              (self.n / self.n_steps)
        self.n += 1
        return res

    def calc_momentum(self) -> None:
        pass

    def update_momentum(self, optimizer: torch.optim.Optimizer) -> float:
        pass

    def on_batch_end(self, i: int, state: DotDict) -> None:
        super().on_batch_end(i, state)
        if self.n == (self.n_steps + 1):
            print("\nEnd of LRFinder")
            state.core.stop_epoch = True


class TBLogger(Callback):
    def __init__(self, logdir: Union[str, Path]) -> None:
        self.logdir = Path(logdir)
        self.writer = None
        self.total_iter = 0
        self.train_iter = 0
        self.val_iter = 0
        self.train_batch_iter = 0
        self.val_batch_iter = 0
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)

    def update_total_iter(self, mode: str) -> None:
        if mode == "train":
            self.train_iter += 1
            self.train_batch_iter += 1
        if mode == "val":
            self.val_iter += 1
            self.val_batch_iter +=1
        self.total_iter += 1

    def on_train_begin(self, state: DotDict) -> None:
        self.train_iter = 0
        self.val_iter = 0
        self.logdir.mkdir(exist_ok=True)
        self.train_writer = SummaryWriter(str(self.logdir / "train"))
        self.val_writer = SummaryWriter(str(self.logdir / "val"))
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict):
        self.train_batch_iter = 0
        self.val_batch_iter = 0

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "train":
            for name, metric in state.core.metrics["train"].items():
                self.train_writer.add_scalar(f"batch/{name}",
                                             float(metric),
                                             global_step=self.total_iter)
                self.train_metrics[name].append(float(metric))

            lr = get_opt_lr(state.core.opt)
            self.train_writer.add_scalar("batch/lr",
                                         float(lr),
                                         global_step=self.train_iter)

            self.update_total_iter(state.core.mode)

        elif state.core.mode == "val":
            for name, metric in state.core.metrics["val"].items():
                self.val_writer.add_scalar(f"batch/{name}",
                                           float(metric),
                                           global_step=self.total_iter)
                self.val_metrics[name].append(float(metric))

            self.update_total_iter(state.core.mode)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if state.core.mode == "train":
            for name, metric in self.train_metrics.items():
                mean = np.mean(metric[-self.train_batch_iter:])
                self.train_writer.add_scalar(f"epoch/{name}",
                                             float(mean),
                                             global_step=epoch)
        if state.core.mode == "val":
            for name, metric in self.val_metrics.items():
                mean = np.mean(metric[-self.val_batch_iter:])  # last epochs vals
                self.val_writer.add_scalar(f"epoch/{name}",
                                           float(mean),
                                           global_step=epoch)

    def on_train_end(self, state: DotDict) -> None:
        self.train_writer.close()
        self.val_writer.close()


class SimpleLossCallback(Callback):
    def __init__(self, target_key: str, preds_key: str) -> None:
        self.target_key = target_key
        self.preds_key = preds_key

    def on_batch_end(self, i: int, state: DotDict) -> None:
        target = state.core.batch[self.target_key]
        preds = state.core.out[self.preds_key]

        state.core.loss = state.core.criterion(preds, target)


class SimpleOptimizerCallback(Callback):
    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "train":
            state.core.opt.zero_grad()
            if state.core.use_fp16:
                with amp.scale_loss(state.core.loss, state.core.opt) as scaled_loss:
                    scaled_loss.backward()
            else:
                state.core.loss.backward()
            state.core.opt.step()


class SimpleSchedulerCallback(Callback):
    def __init__(self,
                 sched: Union[_LRScheduler, ReduceLROnPlateau]) -> None:
        self.metric = "val_loss"
        if isinstance(sched, ReduceLROnPlateau):
            self.when = "on_epoch_end"
        else:
            self.when = "on_epoch_begin"

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        if self.when == "on_epoch_begin" and state.core.mode == "train":
            state.core.sched.step()

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if self.when == "on_epoch_end" and state.core.mode == "train":
            state.core.sched.step(state.core.epoch_metrics[self.metric])


class ProgressBarCallback(Callback):
    def __init__(self) -> None:
        self.running_loss = None

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        self.running_loss = None

        loader = state.core.loader
        if state.core.mode == "train":
            description = f"Epoch {epoch+1}/{epochs}"
            state.core.pbar = get_pbar(loader, description)
        elif state.core.mode == "test":
            description = "Predict"
            state.core.pbar = get_pbar(loader, description)

    def on_batch_end(self, i: int, state: DotDict) -> None:

        if state.core.mode == "train":
            if not self.running_loss:
                self.running_loss = to_numpy(state.core.loss)

            self.running_loss = exp_weight_average(state.core.loss,
                                                   self.running_loss)
            postfix = {"loss": f"{self.running_loss:.4f}"}
            state.core.pbar.set_postfix(postfix)
            state.core.pbar.update()
        elif state.core.mode == "test":
            state.core.pbar.update()

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if state.core.mode == "val":
            metrics = state.core.get("epoch_metrics", {})
            state.core.pbar.set_postfix_str(extend_postfix(state.core.pbar.postfix,
                                                      metrics))
            state.core.pbar.close()
        elif state.core.mode == "test":
            state.core.pbar.close()

class MetricsCallback(Callback):
    def __init__(self,
                 target_key: str,
                 preds_key: str,
                 metrics: Optional[Dict[str, Callable]] = None) -> None:
        self.metrics = metrics or {}
        self.pbar_metrics = None
        self.target_key = target_key
        self.preds_key = preds_key

    def get_metric(self,
                   metric: Callable,
                   target: torch.Tensor,
                   preds: Union[List, torch.Tensor]) -> float:
        # dataparallel workaround
        if isinstance(preds, list):
            preds = torch.cat(preds)
        return metric(target, preds)

    def update_epoch_metrics(self,
                             target: torch.Tensor,
                             preds: Union[List, torch.Tensor]) -> None:

        for name, m in self.metrics.items():
            value = self.get_metric(m, target, preds)
            self.pbar_metrics[name] += value

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        self.pbar_metrics = defaultdict(float)

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "val":
            self.pbar_metrics["val_loss"] += float(to_numpy(state.core.loss))
            self.update_epoch_metrics(target=state.core.batch[self.target_key],
                                      preds=state.core.out[self.preds_key])
        # tb logs
        if state.core.mode != "test" and state.core.do_log:
            state.core.metrics[state.core.mode]["loss"] = float(to_numpy(state.core.loss))
            for name, m in self.metrics.items():
                preds = state.core.out[self.preds_key]
                target = state.core.batch[self.target_key]
                value = self.get_metric(m, target, preds)
                state.core.metrics[state.core.mode][name] = value

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        divider = len(state.core.loader)
        for k in self.pbar_metrics.keys():
            self.pbar_metrics[k] /= divider
        state.core.epoch_metrics = self.pbar_metrics


class PredictionsSaverCallback(Callback):
    def __init__(self,
                 savepath: Optional[Union[str, Path]],
                 preds_key: str) -> None:
        if savepath is not None:
            self.savepath = Path(savepath)
            self.savepath.parent.mkdir(exist_ok=True)
            self.return_array = False
        else:
            self.savepath = None
            self.return_array = True
        self.preds_key = preds_key
        self.preds = []

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "test":
            out = state.core.out[self.preds_key]
            # DataParallelModel workaround
            if isinstance(out, list):
                out = np.concatenate([to_numpy(o) for o in out])
            else:
                out = to_numpy(out)
            self.preds.append(out)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if state.core.mode == "test":
            preds = np.concatenate(self.preds)
            if self.return_array:
                state.core.preds = preds
            else:
                np.save(self.savepath, preds)
            self.preds = []


class CheckpointSaverCallback(Callback):
    def __init__(self,
                 savedir: str,
                 metric: Optional[str] = None,
                 n_best: int = 3,
                 prefix: Optional[str] = None,
                 mode: str = "min") -> None:
        self.metric = metric or "val_loss"
        self.n_best = n_best
        self.savedir = Path(savedir)
        self.prefix = f"{prefix}." if prefix is not None else "checkpoint."

        if mode not in ["min", "max"]:
            raise ValueError(f"mode should be 'min' or 'max', got {mode}")
        if mode == "min":
            self.maximize = False
        if mode == "max":
            self.maximize = True

        self.best_scores = []

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        # trim best scores
        self.best_scores = self.best_scores[:epochs]

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if state.core.mode == "val":
            score_val = state.core.epoch_metrics[self.metric]
            score_name = f"{self.prefix}{epoch + 1}.h5"
            score = (score_val, score_name)
            sorted_scores = sorted(self.best_scores + [score],
                                   reverse=self.maximize)
            self.best_scores = sorted_scores[:self.n_best]
            # set_trace()
            if score_name in (s[1] for s in self.best_scores):
                state.core.checkpoint = f"{self.savedir / score_name}"
                # remove worst checkpoint
                if len(sorted_scores) > self.n_best:
                    # set_trace()
                    Path(f"{self.savedir / sorted_scores[-1][1]}").unlink()

    def on_train_end(self, state: DotDict) -> None:
        best_cp = self.savedir / self.best_scores[0][1]
        shutil.copy(str(best_cp), f"{self.savedir}/{self.prefix}best.h5")
        print(f"\nCheckpoint\t{self.metric or 'val_loss'}")
        for score in self.best_scores:
            print(f"{self.savedir/score[1]}\t{score[0]:.6f}")


class EarlyStoppingCallback(Callback):
    def __init__(self,
                 patience: int,
                 metric: Optional[str] = None,
                 mode: str = "min",
                 min_delta: int = 0) -> None:
        self.best_score = None
        self.metric = metric or "val_loss"
        self.patience = patience
        self.num_bad_epochs = 0
        self.is_better = None

        if mode not in ["min", "max"]:
            raise ValueError(f"mode should be 'min' or 'max', got {mode}")
        if mode == "min":
            self.is_better = lambda score, best: score <= (best - min_delta)
        if mode == "max":
            self.is_better = lambda score, best: score >= (best - min_delta)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if state.core.mode == "val":
            score = state.core.epoch_metrics[self.metric]
            if self.best_score is None:
                self.best_score = score
            if self.is_better(score, self.best_score):
                self.num_bad_epochs = 0
                self.best_score = score
            else:
                self.num_bad_epochs += 1

            if self.num_bad_epochs >= self.patience:
                state.core.stop_train = True


class DebuggerCallback(Callback):
    def __init__(self, when: List[str], modes: List[str]) -> None:
        self.when = when
        self.modes = modes

    def on_batch_begin(self, i: int, state: DotDict) -> None:
        if "on_batch_begin" in self.when:
            if state.core.mode == "train" and "train" in self.modes:
                set_trace()
            if state.core.mode == "val" and "val" in self.modes:
                set_trace()
            if state.core.mode == "test" and "test" in self.modes:
                set_trace()

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if "on_batch_end" in self.when:
            if state.core.mode == "train" and "train" in self.modes:
                set_trace()
            if state.core.mode == "val" and "val" in self.modes:
                set_trace()
            if state.core.mode == "test" and "test" in self.modes:
                set_trace()

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict) -> None:
        if "on_epoch_begin" in self.when:
            if state.core.mode == "train" and "train" in self.modes:
                set_trace()
            if state.core.mode == "val" and "val" in self.modes:
                set_trace()
            if state.core.mode == "test" and "test" in self.modes:
                set_trace()

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:
        if "on_epoch_end" in self.when:
            if state.core.mode == "train" and "train" in self.modes:
                set_trace()
            if state.core.mode == "val" and "val" in self.modes:
                set_trace()
            if state.core.mode == "test" and "test" in self.modes:
                set_trace()

    def on_train_begin(self, state: DotDict) -> None:
        if "on_train_begin" in self.when:
            set_trace()

    def on_train_end(self, state: DotDict) -> None:
        if "on_train_end" in self.when:
            set_trace()



import threading

import torch
from torch.nn.modules import Module
from torch.nn.parallel.scatter_gather import scatter_kwargs
from torch.nn.parallel.replicate import replicate
from torch.nn.parallel.parallel_apply import parallel_apply
from torch.autograd import Function
import torch.cuda.comm as comm
from torch.nn.parallel._functions import Broadcast


__all__ = ['DataParallelModel', 'DataParallelCriterion']


class Reduce(Function):
    @staticmethod
    def forward(ctx, *inputs):
        ctx.target_gpus = [inputs[i].get_device() for i in range(len(inputs))]
        inputs = sorted(inputs, key=lambda i: i.get_device())
        return comm.reduce_add(inputs)

    @staticmethod
    def backward(ctx, *gradOutput):
        return Broadcast.apply(ctx.target_gpus, gradOutput[0])


class DataParallelModel(Module):
    def __init__(self, module, device_ids=None, output_device=None, dim=0):
        super(DataParallelModel, self).__init__()

        if not torch.cuda.is_available():
            self.module = module
            self.device_ids = []
            return

        if device_ids is None:
            device_ids = list(range(torch.cuda.device_count()))
        if output_device is None:
            output_device = device_ids[0]

        self.dim = dim
        self.module = module
        self.device_ids = device_ids
        self.output_device = output_device

        if len(self.device_ids) == 1:
            self.module.cuda(device_ids[0])

    def forward(self, *inputs, **kwargs):
        if not self.device_ids:
            return self.module(*inputs, **kwargs)
        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        if len(self.device_ids) == 1:
            return self.module(*inputs[0], **kwargs[0])
        replicas = self.replicate(self.module, self.device_ids[:len(inputs)])
        outputs = self.parallel_apply(replicas, inputs, kwargs)
        return outputs

    @staticmethod
    def replicate(module, device_ids):
        return replicate(module, device_ids)

    def scatter(self, inputs, kwargs, device_ids):
        return scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)

    def parallel_apply(self, replicas, inputs, kwargs):
        return parallel_apply(replicas, inputs, kwargs, self.device_ids[:len(replicas)])


class DataParallelCriterion(Module):
    def __init__(self, module, device_ids=None, output_device=None, dim=0):
        super(DataParallelCriterion, self).__init__()

        if not torch.cuda.is_available():
            self.module = module
            self.device_ids = []
            return

        if device_ids is None:
            device_ids = list(range(torch.cuda.device_count()))
        if output_device is None:
            output_device = device_ids[0]

        self.dim = dim
        self.module = module
        self.device_ids = device_ids
        self.output_device = output_device

        if len(self.device_ids) == 1:
            self.module.cuda(device_ids[0])

    def forward(self, inputs, *targets, **kwargs):
        # input should be already scattered
        # scattering the targets instead
        targets, kwargs = self.scatter(targets, kwargs, self.device_ids)
        if len(self.device_ids) == 1:
            return self.module(inputs, *targets[0], **kwargs[0])
        inputs = [(input,) for input in inputs]
        replicas = self.replicate(self.module, self.device_ids[:len(inputs)])
        outputs = self.parallel_apply(replicas, inputs, targets, kwargs)
        return self.gather(outputs, self.output_device)

    @staticmethod
    def replicate(module, device_ids):
        return replicate(module, device_ids)

    def scatter(self, inputs, kwargs, device_ids):
        return scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)

    def parallel_apply(self, replicas, inputs, targets, kwargs):
        return criterion_parallel_apply(
            replicas, inputs, targets, kwargs, self.device_ids[:len(replicas)])

    @staticmethod
    def gather(outputs, output_device):
        if not len(outputs[0].shape):
            outputs = [output.unsqueeze(0) for output in outputs]
        return Reduce.apply(*outputs) / len(outputs)


def criterion_parallel_apply(modules, inputs, targets, kwargs_tup=None, devices=None):
    assert len(modules) == len(inputs)
    assert len(targets) == len(inputs)
    if kwargs_tup is not None:
        assert len(modules) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(modules)
    if devices is not None:
        assert len(modules) == len(devices)
    else:
        devices = [None] * len(modules)

    lock = threading.Lock()
    results = {}
    grad_enabled = torch.is_grad_enabled()

    def _worker(i, module, input, target, kwargs, device=None):
        torch.set_grad_enabled(grad_enabled)
        if device is None:
            device = get_a_var(input).get_device()
        try:
            with torch.cuda.device(device):
                output = module(*input, *target, **kwargs)
            with lock:
                results[i] = output
        except Exception as e:
            with lock:
                results[i] = e

    if len(modules) > 1:
        threads = [threading.Thread(target=_worker,
                                    args=(i, module, input, target, kwargs, device),
                                    )
                   for i, (module, input, target, kwargs, device) in
                   enumerate(zip(modules, inputs, targets, kwargs_tup, devices))]

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
    else:
        _worker(0, modules[0], inputs[0], targets[0], kwargs_tup[0], devices[0])

    outputs = []
    for i in range(len(inputs)):
        output = results[i]
        if isinstance(output, Exception):
            raise output
        outputs.append(output)
    return outputs


def get_a_var(obj):
    if isinstance(obj, torch.tensor):
        return obj

    if isinstance(obj, list) or isinstance(obj, tuple):
        for result in map(get_a_var, obj):
            if isinstance(result, torch.tensor):
                return result
    if isinstance(obj, dict):
        for result in map(get_a_var, obj.items()):
            if isinstance(result, torch.tensor):
                return result
    return None


from typing import Optional, Type

import torch
from torch import nn


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size()[0], -1)


# https://github.com/fastai/fastai/blob/e8c855ac70d9d1968413a75c7e9a0f149d28cab3/fastai/layers.py#L171
class AdaptiveConcatPool2d(nn.Module):
    "Layer that concats `AdaptiveAvgPool2d` and `AdaptiveMaxPool2d`."
    def __init__(self,
                 size: Optional[int] = None):
        "Output will be 2*size or 2 if size is None"
        super().__init__()
        size = size or 1
        self.ap = nn.AdaptiveAvgPool2d(size)
        self.mp = nn.AdaptiveMaxPool2d(size)

    def forward(self, x: Type[torch.Tensor]) -> Type[torch.Tensor]:
        return torch.cat([self.mp(x), self.ap(x)], 1)



from collections import namedtuple
from typing import Dict, Callable, Optional
import pandas as pd

from torch.utils.data import Dataset
from torchvision.transforms import Compose


class DataKek(Dataset):
    """The easiest way to store the dataset info is pandas DataFrame.
    So DataKek gets a DataFrame as a source of data"""
    def __init__(self,
                 df: pd.DataFrame,
                 reader_fn: Callable,
                 transforms: Optional[Compose] = None) -> None:
        # transform df to list
        self.data = list(df.values)  # for multiprocessing
        self.df_columns = df.columns
        self.reader_fn = reader_fn
        self.transforms = transforms

    def __getitem__(self, ind: int) -> Dict:
        data_dict = {col: v for col, v in zip(self.df_columns, self.data[ind])}
        datum = self.reader_fn(ind, data_dict)
        if self.transforms is not None:
            datum = self.transforms(datum)
        return datum

    def __len__(self) -> int:
        return len(self.data)


DataOwner = namedtuple("DataOwner", ["train_dl", "val_dl", "test_dl"])



from typing import Callable, Dict, Tuple

import numpy as np
import torch
from torchvision.transforms import Normalize


IMAGENET_STATS = (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)


class Transformer:
    def __init__(self, key: str, transform_fn: Callable) -> None:
        self.key = key
        self.transform_fn = transform_fn

    def __call__(self,
                 datum: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        datum = datum
        datum[self.key] = self.transform_fn(datum[self.key])
        return datum


def to_torch(scale_factor: float = 255.) -> Callable:
    return lambda x: torch.from_numpy(
        x.astype(np.float32) / scale_factor).permute(2, 0, 1)


def normalize(stats: Tuple = IMAGENET_STATS) -> Normalize:
    return Normalize(*stats)



from typing import List, Tuple, Type, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


class FocalLoss(nn.Module):
    def __init__(self,
                 alpha=Union[Tuple[float, int], List],
                 gamma: int = 0,
                 size_average: bool = True) -> None:
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha,(float,int)): self.alpha = torch.Tensor([alpha,1-alpha])
        if isinstance(alpha,list): self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self,
                input: Type[torch.Tensor],
                target: Type[torch.Tensor]) -> Type[torch.Tensor]:
        if input.dim()>2:
            input = input.view(input.size(0),input.size(1),-1)  # N,C,H,W => N,C,H*W
            input = input.transpose(1,2)    # N,C,H*W => N,H*W,C
            input = input.contiguous().view(-1,input.size(2))   # N,H*W,C => N*H*W,C
        target = target.view(-1,1)

        logpt = F.log_softmax(input)
        logpt = logpt.gather(1,target)
        logpt = logpt.view(-1)
        pt = Variable(logpt.data.exp())

        if self.alpha is not None:
            if self.alpha.type()!=input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0,target.data.view(-1))
            logpt = logpt * Variable(at)

        loss = -1 * (1-pt)**self.gamma * logpt
        if self.size_average: return loss.mean()
        else: return loss.sum()


"""MIT License

Copyright (c) 2017 carwin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software."""


from typing import Type

import torch

from sklearn.metrics import accuracy_score, roc_auc_score


def accuracy(target: torch.Tensor,
             preds: torch.Tensor) -> float:
    target = target.cpu().detach().numpy()
    preds = preds.cpu().detach().numpy().argmax(axis=1)
    return accuracy_score(target, preds)

def bce_accuracy(target: torch.Tensor,
                 preds: torch.Tensor,
                 thresh: bool = 0.5) -> float:
    target = target.cpu().detach().numpy()
    preds = (torch.sigmoid(preds).cpu().detach().numpy() > thresh).astype(int)
    return accuracy_score(target, preds)
  
def roc_auc(target: torch.Tensor,
                 preds: torch.Tensor) -> float:
    target = target.cpu().detach().numpy()
    preds = torch.sigmoid(preds).cpu().detach().numpy()
    return roc_auc_score(target, preds)



from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, List, Tuple, Type, Dict, Union, Optional
import warnings

try:
    from apex import amp
except ImportError as e:
    warnings.warn(f"Error '{e}'' during importing apex library. To use mixed precison"
                  " you should install it from https://github.com/NVIDIA/apex")

import numpy as np

import torch
from torch.utils.data import DataLoader
from torch.optim import SGD
from torch.nn.parallel import DistributedDataParallel

from .callbacks import Callback, Callbacks, ProgressBarCallback, \
    PredictionsSaverCallback, OneCycleLR, SimpleLossCallback, MetricsCallback, \
    TBLogger, LRFinder, CheckpointSaverCallback, SimpleSchedulerCallback, \
    EarlyStoppingCallback, SimpleOptimizerCallback
from .data import DataOwner
from .parallel import DataParallelCriterion, DataParallelModel
from .utils import DotDict, freeze_to, freeze, unfreeze, load_state_dict, \
    plot_tensorboard_log


class Keker:
    """ The class serving the whole train-val-predict process.

    Args:
        model: The neural network for train/val/predict.
        dataowner: The namedtuple container of train/val/test dataloaders.
        criterion: The loss function or the dict {'name': loss function}
            in case of multiple loss setup. If multiple loss is using,
            loss_cb should be provided.
            (ex. : torch.nn.CrossEntropyLoss(),
            {"ce": torch.nn.CrossEntropyLoss(), "bce": torch.nn.BCE()})
        tarket_key: The target/label key for batch-dict from dataloader.
            The dataloader returns batch as a dict on each iteration,
            that contains input data and target labels. This is key is for
            access to target labels in this dict.
        preds_key: The key for dict from self.step() functions.
            The self.step() function returns dict of predictions on each batch.
            This key is for access to predictions in this dict.
            This attribute is optional in default behavior but coulb be used
            in case of using custom callbacks.
        metrics: {"name": metric_function} dict, that contains callable metrics
            for calculating. The metric takes prediction and target
            tensors as parameters, and returns float.
            For examples see kekas.metrics module.
        opt: pytorch Optimizer class (ex. torch.optim.SGD, torch.optm.Adam, etc)
            This optimizer will be used as default during training.
            Default optimizer is torch.optim.SGD.
        opt_params: The kwargs dict for optimizer initialization.
            It should contain any optimizer param you want EXCEPT learning rate,
            learing rate is specified in self.kek* methods.
        device: The torch.device when you want to put your model
        step_fn: The function that will be called at every batch of your data.
            Take a `self` as a parameter. In this function you define
            what you do with your batch. You get access to batch through
            self._state.core.batch object. Batch is a dict returned from your
            dataloader, and its key and values are specified in reader_function
            for DataKek.
            Return a dict that should contain batch predictions with access
            by `preds_key`.
            For example see self.default_step method and example ipynbs.
        loss_cb: The Callback for loss calculation. Define how to calculate loss
            and store loss value in self._state.core.loss attr.
            Default loss callback is SimpleLossCallback.
            For examples see kekas.callbacks.
        opt_cb: The Callback for optimizer applying.
            Default optimizer callback is SimpleOptimizerCallback.
            For examples see kekas.callbacks.
        callbacks: custom Callbacks thet will be applied before the core ones.
            For examples see example ipynbs.
    """
    def __init__(self,
                 model: torch.nn.Module,
                 dataowner: Optional[DataOwner] = None,
                 criterion: Optional[Union[torch.nn.Module,
                                           Dict[str, torch.nn.Module]]] = None,
                 target_key: str = "label",
                 preds_key: str = "preds",
                 metrics: Optional[Dict[str, Callable]] = None,
                 opt: Optional[Type[torch.optim.Optimizer]] = None,
                 opt_params: Optional[Dict] = None,
                 device: Optional[torch.device] = None,
                 step_fn: Optional[Type[Callback]] = None,
                 loss_cb: Optional[Type[Callback]] = None,
                 opt_cb: Optional[Type[Callback]] = None,
                 callbacks: Optional[Union[List, Callbacks]] = None) -> None:

        # The state is an object that stores many variables and represents
        # the state of your train-val-predict pipeline. state passed to every
        # callback call.
        # You can use it as a container for your custom variables, but
        # DO NOT USE the 'core' attrubute, because all kekers variables are
        # there
        #
        self.state = DotDict()
        self.state.core = DotDict()

        self.state.core.model = model
        self.state.core.dataowner = dataowner or DataOwner(None, None, None)

        self.target_key = target_key
        self.preds_key = preds_key

        self.state.core.criterion = criterion

        self.state.core.parallel = False
        if torch.cuda.device_count() > 1:
            self.state.core.model = DataParallelModel(self.state.core.model)
            if self.state.core.criterion is not None:
                if isinstance(self.state.core.criterion, dict):
                    self.state.core.criterion = {
                        k: DataParallelCriterion(v) for k, v
                        in self.state.core.criterion.items()
                    }
                else:
                    self.state.core.criterion = DataParallelCriterion(
                        self.state.core.criterion
                    )
            self.state.core.parallel = True

        self.opt = opt or SGD
        self.opt_params = opt_params or {}
        self.device = device or torch.device("cuda" if
                                             torch.cuda.is_available()
                                             else "cpu")
        self.state.core.model.to(self.device)

        self.step_fn = step_fn or self.default_step_fn

        # the core callbacks for train-val-predict are determined here.
        # the order of callbacks is matters!
        if loss_cb is not None:
            loss_cb = loss_cb(target_key, self.preds_key)
        else:
            loss_cb = SimpleLossCallback(target_key, self.preds_key)

        opt_cb = opt_cb or SimpleOptimizerCallback()
        metrics_cb = MetricsCallback(target_key, self.preds_key, metrics)

        callbacks = callbacks or []
        self.core_callbacks = callbacks + [loss_cb,
                                           metrics_cb,
                                           opt_cb,
                                           ProgressBarCallback()]
        callbacks = self.core_callbacks[:]

        self.callbacks = Callbacks(callbacks)

        self.state.core.checkpoint = ""

        # The number of batch in dataloader for iteration stop,
        # determined in self.kek* methods.
        self.state.core.stop_iter = None

        # flag for train stop after batch.
        self.state.core.stop_epoch = False

        # flag for stop the whole train, used for early stopping.
        self.state.core.stop_train = False

        # The scheduler attribute. Scheduler is determined in self.kek* methods.
        self.state.core.sched = None

        # Flag for logger callback
        self.state.core.do_log = False

        # Predictions variable
        self.state.core.preds = None

        # FP16 flag
        self.state.core.use_fp16 = False

    def kek(self,
            lr: float,
            epochs: int,
            skip_val: bool = False,
            opt: Optional[Type[torch.optim.Optimizer]] = None,
            opt_params: Optional[Dict] = None,
            sched: Optional[Callable] = None,
            sched_params: Optional[Dict] = None,
            stop_iter: Optional[int] = None,
            logdir: Optional[Union[str, Path]] = None,
            cp_saver_params: Optional[Dict] = None,
            early_stop_params: Optional[Dict] = None) -> None:
        """Kek your model to the moon!

        Conducts a standard train-val procedure with several options for
        customization.

        Args:
            lr: learining rate
            epochs: number of epochs to train
            opt: torch optimizer. If specified, than specified optimizer will be
                used for this train-val procedure, else the default one.
            opt_params: The kwargs dict for custom optimizer initialization.
                It should contain any opt params you want EXCEPT learning rate,
                Can be defined even for default optimizer.
            sched: optional pytorch scheduler class. If specified, sched_params
                must be specified too.
                Ex: torch.optim.lr_scheduler.StepLR.
            sched_params: kwargs dict parameters for scheduler
            stop_iter: number of batch when you want to end an epoch
            logdir: If provided, the TBLogger will be created and tensorboard
                logs will be written in this directory.
                For more info see kekas.callbacks.TBLogger and example ipynb's.
            cp_saver_params: kwargs dict parameters for CheckpointSaverCallback.
                If provided, then a CheckpointSaverCallback will be created.
                For more info see kekas.callbacks.CheckpointSaverCallback
                and example ipynb's.
            early_stop_params: kwargs dict parameters for EarlyStoppingCallback.
                If provided, then a EarlyStoppingCallback will be created.
                For more info see kekas.callbacks.EarlyStoppingCallback
                and example ipynb's.
        """

        # check if criterion exists
        if self.state.core.criterion is None:
            raise Exception("Keker needs criterion. "
                            "Reinitialize Keker with one.")

        # check if dataowner exists
        if self.state.core.dataowner.train_dl is None:
            raise Exception("Keker needs Dataowner. "
                            "Reinitialize Keker with one.")

        if stop_iter:
            self.stop_iter = stop_iter

        # save callbacks
        callbacks = self.callbacks

        opt = opt or self.opt
        opt_params = opt_params or self.opt_params
        params = (p for p in self.state.core.model.parameters() if p.requires_grad)
        self.state.core.opt = opt(params=params, lr=lr, **opt_params)

        if self.state.core.use_fp16:
            _, self.state.core.opt = amp.initialize(self.state.core.model,
                                                    self.state.core.opt,
                                                    opt_level="O1",
                                                    verbosity=0)

        if sched:
            sched_params = sched_params or {}
            self.state.core.sched = sched(optimizer=self.state.core.opt, **sched_params)
            sched_cb = SimpleSchedulerCallback(sched=self.state.core.sched)
            self.callbacks = Callbacks(self.callbacks.callbacks + [sched_cb])

        if logdir:
            self.state.core.do_log = True
            self.state.core.metrics = defaultdict(dict)
            tboard_cb = TBLogger(logdir)
            self.callbacks = Callbacks(self.callbacks.callbacks + [tboard_cb])

        cp_saver_params = cp_saver_params or {}
        if cp_saver_params:
            cp_saver_cb = CheckpointSaverCallback(**cp_saver_params)
            self.callbacks = Callbacks(self.callbacks.callbacks + [cp_saver_cb])

        early_stop_params = early_stop_params or {}
        if early_stop_params:
            early_stop_cb = EarlyStoppingCallback(**early_stop_params)
            self.callbacks = Callbacks(self.callbacks.callbacks + [early_stop_cb])

        # try-finally to properly close progress bar and restore callbacks
        try:
            self.callbacks.on_train_begin(self.state)

            for epoch in range(epochs):
                self.set_mode("train")
                self._run_epoch(epoch, epochs)

                if not skip_val:
                    self.set_mode("val")
                    self._run_epoch(epoch, epochs)

                if self.state.core.stop_train:
                    self.state.core.stop_train = False
                    print(f"Early stopped on {epoch + 1} epoch")
                    break

            self.callbacks.on_train_end(self.state)
        finally:
            self.state.core.pbar.close()
            self.state.core.do_log = False
            self.callbacks = callbacks

    def kek_one_cycle(self,
                      max_lr: float,
                      cycle_len: int,
                      momentum_range: Tuple[float, float] = (0.95, 0.85),
                      div_factor: float = 25,
                      increase_fraction: float = 0.3,
                      opt: Optional[Type[torch.optim.Optimizer]] = None,
                      opt_params: Optional[Dict] = None,
                      logdir: Optional[Union[str, Path]] = None,
                      cp_saver_params: Optional[Dict] = None,
                      early_stop_params: Optional[Dict] = None) -> None:
        """Kek your model to the moon with One Cycle policy!

        Conducts a one-cycle train-val procedure with several options for
        customization.
        For info about One Cycle policy please see:
        https://arxiv.org/abs/1803.09820
        https://sgugger.github.io/the-1cycle-policy.html

        Args:
            max_lr: the maximum learning rate that will be achieved during
                training process
            cycle_len: the number of full passes through the training dataset.
                It is quite similar to epochs number
            momentum_range: the range of optimizers momentum changes
            div_factor: is equal to max_lr / min_lr during one-cycle training
            increase_fraction: the fraction of the whole iterations during which
                the learning rate will increase
            opt: torch optimizer. If specified, than specified optimizer will be
                used for this train-val procedure, else the default one.
            opt_params: The kwargs dict for custom optimizer initialization.
                It should contain any opt params you want EXCEPT learning rate,
                Can be defined even for default optimizer.
            logdir: If provided, the TBLogger will be created and tensorboard
                logs will be written in this directory.
                For more info see kekas.callbacks.TBLogger and example ipynb's.
            cp_saver_params: kwargs dict parameters for CheckpointSaverCallback.
                If provided, then a CheckpointSaverCallback will be created.
                For more info see kekas.callbacks.CheckpointSaverCallback
                and example ipynb's.
            early_stop_params: kwargs dict parameters for EarlyStoppingCallback.
                If provided, then a EarlyStoppingCallback will be created.
                For more info see kekas.callbacks.EarlyStoppingCallback
                and example ipynb's.
        """

        callbacks = self.callbacks

        # temporarily add OneCycle callback
        len_loader = len(self.state.core.dataowner.train_dl)
        one_cycle_cb = OneCycleLR(max_lr, cycle_len, len_loader,
                                  momentum_range, div_factor, increase_fraction)

        try:
            self.callbacks = Callbacks(callbacks.callbacks + [one_cycle_cb])

            self.kek(lr=max_lr,
                     epochs=cycle_len,
                     opt=opt,
                     opt_params=opt_params,
                     logdir=logdir,
                     cp_saver_params=cp_saver_params,
                     early_stop_params=early_stop_params)
        finally:
            # set old callbacks without OneCycle
            self.callbacks = callbacks

    def kek_lr(self,
               final_lr: float,
               logdir: Union[str, Path],
               init_lr: float = 1e-6,
               n_steps: Optional[int] = None,
               opt: Optional[Type[torch.optim.Optimizer]] = None,
               opt_params: Optional[Dict] = None) -> None:
        """Help you kek your model to the moon by finding "optimal" lr!

        Conducts the learning rate find procedure.
        For info please see:
        https://arxiv.org/abs/1803.09820
        https://sgugger.github.io/how-do-you-find-a-good-learning-rate.html

        Args:
            final_lr: the learning rate at the end of the lr find process
            logdir: the directory for tensorboard logs, that will allow you
                analyze the loss dynamic.
            init_lr: the learning rate at the start of the lr find process
            n_steps: the number of iterations of lr find process. If provided
                finding process will stop at this iteration, else lr find
                process will last one epoch
            opt: torch optimizer. If specified, than specified optimizer will be
                used for this train-val procedure, else the default one.
            opt_params: The kwargs dict for custom optimizer initialization.
                It should contain any opt params you want EXCEPT learning rate,
                Can be defined even for default optimizer.
        """

        logdir = Path(logdir)
        logdir.mkdir(exist_ok=True)
        tmp_cp = logdir / "tmp.h5"
        self.save(tmp_cp)

        len_loader = len(self.state.core.dataowner.train_dl)
        n_steps = n_steps if n_steps is not None else len_loader
        n_epochs = max(1, int(np.ceil(n_steps / len_loader)))

        callbacks = self.callbacks

        try:
            lrfinder_cb = LRFinder(final_lr=final_lr,
                                   init_lr=init_lr,
                                   n_steps=n_steps)

            self.callbacks = Callbacks(self.core_callbacks + [lrfinder_cb])
            self.kek(lr=init_lr, epochs=n_epochs, skip_val=True, logdir=logdir,
                     opt=opt, opt_params=opt_params)
        finally:
            self.callbacks = callbacks
            self.load(tmp_cp)
            tmp_cp.unlink()

    def _run_epoch(self,
                   epoch: int,
                   epochs: int) -> None:
        """Run one epoch of train-val procedure

        Args:
            epoch: number of the current epoch
            epochs: total number of epochs
        """
        self.callbacks.on_epoch_begin(epoch, epochs, self.state)

        with torch.set_grad_enabled(self.is_train):
            for i, batch in enumerate(self.state.core.loader):
                self.callbacks.on_batch_begin(i, self.state)

                self.state.core.batch = self.to_device(batch)

                self.state.core.out = self.step()

                self.callbacks.on_batch_end(i, self.state)

                if (self.state.core.stop_iter and self.state.core.mode == "train"
                        and i == self.state.core.stop_iter - 1):
                    # break only in train mode and if early stop is set
                    self.state.core.stop_epoch = True

                if self.state.core.stop_epoch:
                    self.state.core.stop_epoch = False
                    # st()
                    break

        self.callbacks.on_epoch_end(epoch, self.state)

        if self.state.core.checkpoint:
            self.save(self.state.core.checkpoint)
            self.state.core.checkpoint = ""

    @staticmethod
    def default_step_fn(model: torch.nn.Module,
                        batch: torch.Tensor) -> torch.Tensor:
        """Determine what your model will do with your data.

        Args:
            model: the pytorch module to pass input in
            batch: the batch of data from the DataLoader

        Returns:
            The models forward pass results
        """
        inp = batch["image"]
        return model(inp)

    def step(self) -> Dict[str, torch.Tensor]:
        """The step function that calls each iteration.
        Wraps the self.step_fn to provide a dict of predictions

        Returns:
            the dict that contains prediction tensor for the batch.
        """
        preds = self.step_fn(model=self.state.core.model,
                             batch=self.state.core.batch)

        return {self.preds_key: preds}

    def predict(self,
                savepath: Optional[Union[str, Path]] = None
                ) -> Union[None, np.ndarray]:
        """Infer the model on test dataloader and saves prediction as numpy array

        Args:
            savepath: the directory to save predictions. If not passed,
                the method will return a numpy array of predictions.
        """
        return self.predict_loader(loader=self.state.core.dataowner.test_dl,
                                   savepath=savepath)

    def predict_loader(self,
                       loader: DataLoader,
                       savepath: Optional[Union[str, Path]] = None
                       ) -> Union[None, np.ndarray]:
        """Infer the model on dataloader and saves prediction as numpy array

        Args:
            loader: the dataloader for generating predictions
            savepath: the directory to save predictions. If not passed,
                the method will return a numpy array of predictions.
        """
        callbacks = self.callbacks

        tmp_callbacks = Callbacks([ProgressBarCallback(),
                                   PredictionsSaverCallback(savepath,
                                                            self.preds_key)])

        self.callbacks = tmp_callbacks

        self.state.core.mode = "test"
        self.state.core.loader = loader
        self.state.core.model.eval()
        with torch.set_grad_enabled(False):
            self._run_epoch(1, 1)

        self.callbacks = callbacks

        preds = self.state.core.preds
        self.state.core.preds = None
        return preds

    def predict_tensor(self,
                       tensor: Type[torch.Tensor],
                       to_numpy: bool = False) -> Union[Type[torch.Tensor],
                                                        np.ndarray]:
        """Infer the model on one torch Tensor.

        Args:
            tensor: torch tensor to predict on.
                Should has [batch_size, *(one_sample_shape)] shape
            to_numpy: if True, converts predictions to numpy array

        Returns:
            Predictions on input tensor.
        """
        tensor = tensor.to(self.device)
        with torch.set_grad_enabled(False):
            self.set_mode("test")
            preds = self.state.core.model(tensor)

            # dataparallel workaround
            if isinstance(preds, list):
                preds = torch.cat(preds)
        if to_numpy:
            preds = preds.cpu().numpy()
        return preds

    def predict_array(self,
                      array: np.ndarray,
                      to_numpy: bool = False) -> Union[Type[torch.Tensor],
                                                       np.ndarray]:
        """Infer the model on one numpy array.

        Args:
            array: numpy array to predict on.
                Should has [batch_size, *(one_sample_shape)] shape
            to_numpy: if True, converts predictions to numpy array

        Returns:
            Predictions on input tensor.
        """
        tensor = torch.from_numpy(array)
        return self.predict_tensor(tensor, to_numpy)

    def TTA(self,
            loader: DataLoader,
            tfms: Union[List, Dict],
            savedir: Union[str, Path],
            prefix: str = "preds") -> None:
        """Conduct the test-time augmentations procedure.

        Create predictions for each set of provided transformations and saves
        each prediction in savedir as a numpy arrays.

        Args:
            loader: loader to predict
            tfms: the list with torchvision.transforms or
                  the dict with {"name": torchvision.transforms} pairs.
                  List indexes or dict keys will be used for generating
                  predictions names.
            savedir: the directory to save predictions
            prefix: the prefix for predictions files names
        """
        if isinstance(tfms, dict):
            names = [f"{prefix}_{k}.npy" for k in tfms]
            tfms = tfms.values()
        elif isinstance(tfms, list):
            names = [f"{prefix}_{i}.npy" for i in range(len(tfms))]
        else:
            raise ValueError(f"Transforms should be List or Dict, "
                             f"got {type(tfms)}")

        default_tfms = loader.dataset.transforms
        for name, tfm in zip(names, tfms):
            loader.dataset.transforms = tfm
            savepath = Path(savedir) / name
            self.predict_loader(loader, savepath)
        loader.dataset.transforms = default_tfms

    def save(self, savepath: Union[str, Path]) -> None:
        """Save models state dict on the specified path.

        Args:
            savepath: the path to save the state dict.
        """
        savepath = Path(savepath)
        savepath.parent.mkdir(exist_ok=True)
        torch.save(self.state.core.model.state_dict(), savepath)

    def load(self,
             loadpath: Union[str, Path],
             skip_wrong_shape: bool = False) -> None:
        """Loads models state dict from the specified path.

        Args:
            loadpath: the path from which the state dict will be loaded.
            skip_wrong_shape: If False, will raise an exception if checkpoints
                weigths shape doesn't match models weights shape.
                If True, will skip unmatched weights and load only matched.
        """
        loadpath = Path(loadpath)
        checkpoint = torch.load(loadpath,
                                map_location=lambda storage, loc: storage)

        # workaround DataParallelModel
        if not isinstance(self.state.core.model, DataParallelModel) \
                and "module." in list(checkpoint.keys())[0]:
            # [7:] is to skip 'module.' in group name
            checkpoint = {k[7:]: v for k, v in checkpoint.items()}

        load_state_dict(model=self.state.core.model,
                        state_dict=checkpoint,
                        skip_wrong_shape=skip_wrong_shape)

    def to_device(self,
                  batch: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Moves tensors in batch to self.device.

        Args:
            batch: the batch dict.

        Returns:
            The batch dict with tensors on self.device.
        """
        res = {}
        for k, v in batch.items():
            if isinstance(v, dict):
                v = self.to_device(v)
            else:
                if hasattr(v, "to"):
                    v = v.to(self.device)
            res[k] = v

        return res

    def to_fp16(self):
        """Use NVIDIA apex library for mixed precision training.
        After calling this method, all operations will be used in mixed precision.

        Returns:
            self
        """
        self.state.core.model = amp.initialize(self.state.core.model,
                                               opt_level="O1",
                                               verbosity=0)
        self.state.core.use_fp16 = True
        return self

    def set_mode(self, mode: str) -> None:
        """Set the model to train or val and switch dataloaders

        Args:
            mode: 'train', 'val' or 'test', the mode of training procedure.
        """
        if mode == "train":
            self.state.core.model.train()
            self.state.core.loader = self.state.core.dataowner.train_dl
        elif mode == "val":
            self.state.core.model.eval()
            self.state.core.loader = self.state.core.dataowner.val_dl
        elif mode == "test":
            self.state.core.model.eval()
            self.state.core.loader = self.state.core.dataowner.test_dl
        self.state.core.mode = mode

    def freeze_to(self,
                  n: int,
                  freeze_bn: bool = False,
                  model_attr: Optional[str] = None) -> None:
        """Freeze model or model's part till specified layer.

        Args:
            n: the layer number to freeze to
            freeze_bn: if True batchnorm layers will be frozen too
            model_attr: the name of the model attribute if you want to specify
                when you want to freeze layers.
                For examples see example ipynb's.
        """

        module = self.get_model_attr(model_attr)
        freeze_to(module, n, freeze_bn)

    def freeze(self,
               freeze_bn: bool = False,
               model_attr: Optional[str] = None) -> None:
        """Freeze model or model's part till the last layer

        Args:
            freeze_bn: if True batchnorm layers will be frozen too
            model_attr: the name of the model attribute if you want to specify
                when you want to freeze layers.
                For examples see example ipynb's.
        """
        module = self.get_model_attr(model_attr)
        freeze(module, freeze_bn)

    def unfreeze(self,
                 model_attr: Optional[str] = None) -> None:
        """Unfreeze all model or model's part layers.

        Args:
            model_attr: the name of the model attribute if you want to specify
                when you want to freeze layers.
                For examples see example ipynb's.
        """
        module = self.get_model_attr(model_attr)
        unfreeze(module)

    def get_model_attr(self, model_attr: Union[str, None]) -> torch.nn.Module:
        """Get models attribute by name or return the model if name is None.

        Args:
            model_attr: models attribute name to get. If none, than the model
                will be returned.

        Returns:
            The models attribute or the model itself.
        """
        if self.state.core.parallel:
            model = self.state.core.model.module
        else:
            model = self.state.core.model

        if model_attr is not None:
            module = getattr(model, model_attr)
        else:
            module = model
        return module

    def add_callbacks(self, callbacks: List[Callback]) -> None:
        """Add callbacks to the beginning of self.callbacks"""
        self.callbacks = Callbacks(callbacks + self.callbacks.callbacks)

    @staticmethod
    def plot_kek(logdir: Union[str, Path],
                 step: Optional[str] = "batch",
                 metrics: Optional[List[str]] = None,
                 height: Optional[int] = None,
                 width: Optional[int] = None) -> None:
        """Plots your keks results in Jupyter Notebook.

        Args:
            logdir: the logdir that was specified during kek.
            step: 'batch' or 'epoch' - what logs to show: for batches or
                for epochs
            metrics: list of metrics to plot. The loss should be specified as
                'loss', learning rate = 'lr' and other metrics should be
                specified as names in metrics dict that was specified during kek
            height: the height of the whole resulting plot
            width: the width of the whole resulting plot

        """
        assert step in ["batch", "epoch"], f"Step should be either 'batch' or" \
                                           f"'epoch', got '{step}'"
        plot_tensorboard_log(logdir, step, metrics, height, width)

    @staticmethod
    def plot_kek_lr(logdir: Union[str, Path],
                    height: Optional[int] = None,
                    width: Optional[int] = None) -> None:
        """Plots learing rate finding results in Jupyter Notebook
        Args:
            logdir: the logdir that was specified during kek_lr.
            height: the height of the whole resulting plot
            width: the width of the whole resulting plot
        """
        step = "batch"
        metrics = ["loss", "lr"]
        plot_tensorboard_log(logdir, step, metrics, height, width)

    @property
    def is_train(self) -> bool:
        return self.state.core.mode == "train"



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kekas",
    version="0.1.18",
    author="Aleksandr Belskikh",
    author_email="belskikh.aleksandr@gmail.com",
    description="Just another DL library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/belskikh/kekas",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    python_requires=">=3.6.0",
    packages=setuptools.find_packages(),
    install_requires=[
        "pandas>=0.22",
        "numpy>=1.14.6",
        "plotly>=3.6.1",
        "tensorboard>=1.12.2,<2.0",
        "tensorboardX>=1.6",
        "tensorflow>=1.12.0,<2.0",
        "torch>=0.4.1",
        "torchvision>=0.2.1",
        "tqdm>=4.29.1",
        "scikit-learn>=0.20"
    ]
)
